"""
Holds & provides configuration parameters for auviewer.
"""

import json
import logging
from pathlib import Path

from .shared import createEmptyJSONFile

# Get the code root
# TODO(gus): This may not work when once viewer is pip installable
codeRootPathObj = Path(__file__).resolve().parent
# Read in the built-in default interface templates file. This will (and should)
# # raise an exception if the file does not exist.
with (codeRootPathObj / 'static' / 'builtin_templates' / 'interface_templates.json').open() as f:
    bdit = f.read()

# Read in the built-in default project template file. This will (and should)
# raise an exception if the file does not exist.
with (codeRootPathObj / 'static' / 'builtin_templates' / 'project_template.json').open() as f:
    bdpt = f.read()

# Holds all config params
config = {

    ### General settings

    # TODO: Do we still need this section?



    ### Tuning parameters

    # Max number of data points to transmit for a given view
    'M': 3000,

    # The number to multiply by numDownsamples each time in building downsample
    # levels. For example, if M=3000 and stepMultipler=2, the first downsample
    # is 3K intervals, the second 6K, the third 12K, and so forth.
    'stepMultiplier': 2,



    ### Asset locations

    # AUView code root directory
    'codeRootPathObj': codeRootPathObj,

    # Data directory location (pathlib.Path object).
    # NOTE: This is the only config param that has no default value and must
    # be defined.
    'dataPathObj': None,

    # Path to the projects directory (pathlib.Path object).
    'projectsDirPathObj': None,



    ### Web server configuration

    'host': 'localhost',
    'port': 8001,
    'debug': False,

    # Root directory from which the web application is served. Should begin with a
    # slash and end without a slash (in other words, end with a directory name). If
    # there is no root directory, rootWebPath should be empty string. You must also
    # change the corresponding setting in config.js. Examples:
    # ''
    # '/approot'
    # '/app/root'
    'rootWebPath': '',

    # Password encryption key
    'secret_key': 'THISISADEVELOPMENTSECRETKEY!CHANGEMETONEWRANDOMSTRINGFORPRODUCTION!',

    # Mail integration settings object
    'mail': {},



    ### Templates

    # Built-in default interface templates
    'builtinDefaultInterfaceTemplates': bdit,

    # Built-in default project template
    'builtinDefaultProjectTemplate': bdpt,

    # Global default interface templates (will be overwritten upon a data folder being designated)
    'globalDefaultInterfaceTemplates': "{}",

    # Global default project template (will be overwritten upon a data folder being designated)
    'globalDefaultProjectTemplate': "{}",


}

# Flask configuration object
class FlaskConfigClass(object):
    """ Flask application config """

    # In production, this is needed to build the URL that will be used in
    # Flask-User emails. However, it will also make it so that Flask only serves
    # requests from this hostname. See:
    # https://code.luasoftware.com/tutorials/flask/things-you-should-know-about-flask-server-name/
    # SERVER_NAME = '127.0.0.1'

    # Flask-SQLAlchemy settings
    SQLALCHEMY_TRACK_MODIFICATIONS = False  # Avoids SQLAlchemy warning

    CSRF_ENABLED = True

    # Flask-User settings
    USER_APP_NAME = "AUViewer"  # Shown in and email templates and page footers
    USER_ENABLE_CHANGE_USERNAME = False
    USER_ENABLE_CONFIRM_EMAIL = False
    USER_ENABLE_EMAIL = True
    USER_ENABLE_INVITATION = True
    USER_ENABLE_USERNAME = False
    USER_ENABLE_RETYPE_PASSWORD = False
    USER_REQUIRE_INVITATION = True

    # Override default Flask-User URLs with root web path prefix. For documentation,
    # see "URLs" section of https://flask-user.readthedocs.io/en/v0.6/customization.html.
    # These are overridden in serve.py (createApp).
    # USER_CHANGE_PASSWORD_URL = rootWebPath+'/user/change-password'
    # USER_CHANGE_USERNAME_URL = rootWebPath+'/user/change-username'
    # USER_CONFIRM_EMAIL_URL = rootWebPath+'/user/confirm-email/<token>'
    # USER_EMAIL_ACTION_URL = rootWebPath+'/user/email/<id>/<action>'
    # USER_FORGOT_PASSWORD_URL = rootWebPath+'/user/forgot-password'
    # USER_INVITE_URL = rootWebPath+'/user/invite'
    # USER_LOGIN_URL = rootWebPath+'/user/login'
    # USER_LOGOUT_URL = rootWebPath+'/user/logout'
    # USER_MANAGE_EMAILS_URL = rootWebPath+'/user/manage-emails'
    # USER_PROFILE_URL = rootWebPath+'/user/profile'
    # USER_REGISTER_URL = rootWebPath+'/user/register'
    # USER_RESEND_CONFIRM_EMAIL_URL = rootWebPath+'/user/resend-confirm-email'
    # USER_RESET_PASSWORD_URL = rootWebPath+'/user/reset-password/<token>'
    USER_CHANGE_PASSWORD_URL = '/user/change-password'
    USER_CHANGE_USERNAME_URL = '/user/change-username'
    USER_CONFIRM_EMAIL_URL = '/user/confirm-email/<token>'
    USER_EMAIL_ACTION_URL = '/user/email/<id>/<action>'
    USER_FORGOT_PASSWORD_URL = '/user/forgot-password'
    USER_INVITE_URL = '/user/invite'
    USER_LOGIN_URL = '/user/login'
    USER_LOGOUT_URL = '/user/logout'
    USER_MANAGE_EMAILS_URL = '/user/manage-emails'
    USER_PROFILE_URL = '/user/profile'
    USER_REGISTER_URL = '/user/register'
    USER_RESEND_CONFIRM_EMAIL_URL = '/user/resend-confirm-email'
    USER_RESET_PASSWORD_URL = '/user/reset-password/<token>'

    # Override default Flask-User endpoints. For documentation, see "Endpoints"
    # section of https://flask-user.readthedocs.io/en/v0.6/customization.html.
    USER_AFTER_CHANGE_PASSWORD_ENDPOINT = 'index'
    USER_AFTER_CHANGE_USERNAME_ENDPOINT = 'index'
    USER_AFTER_CONFIRM_ENDPOINT = 'index'
    USER_AFTER_FORGOT_PASSWORD_ENDPOINT = 'user.login'
    USER_AFTER_LOGIN_ENDPOINT = 'index'
    USER_AFTER_LOGOUT_ENDPOINT = 'user.login'
    USER_AFTER_REGISTER_ENDPOINT = 'user.login'
    USER_AFTER_RESEND_CONFIRM_EMAIL_ENDPOINT = 'user.login'
    USER_AFTER_RESET_PASSWORD_ENDPOINT = 'user.login'
    USER_INVITE_ENDPOINT = 'index'
    USER_UNCONFIRMED_EMAIL_ENDPOINT = 'user.login'
    USER_UNAUTHENTICATED_ENDPOINT = 'user.login'
    USER_UNAUTHORIZED_ENDPOINT = 'index'

# Load configuration file
def load_config(cp):

    global config

    logging.info(f"Loading config file {cp}.")

    # Unmarshal the config file from JSON into a dict
    with cp.open() as f:
        json_config = json.load(f)

    # The following values are valid and can be pulled from the json config file
    possible_keys = [
        'host',
        'port',
        'debug',

        'rootWebPath',
        'secret_key',
        'mail',
    ]

    # Set/override any valid settings provided in the json config file
    for k in possible_keys:
        if k in json_config:
            config[k] = json_config[k]

# Generate the baseline data directory contents as needed
def scaffold_data_path():

    global config

    # Grab the data folder path object
    p = config['dataPathObj']

    # Scaffold
    (p / 'config').mkdir(exist_ok=True)
    (p / 'database').mkdir(exist_ok=True)
    (p / 'global_templates').mkdir(exist_ok=True)
    (p / 'projects').mkdir(exist_ok=True)
    createEmptyJSONFile(p / 'config' / 'config.json')
    createEmptyJSONFile(p / 'global_templates' / 'interface_templates.json')
    createEmptyJSONFile(p / 'global_templates' / 'project_template.json')

# Sets, prepares, and processes the data path.
def set_data_path(path):

    global config

    # Create a path object
    p = Path(path)

    # Resolve user directory, if necessary
    if path.startswith('~'):
        p = p.expanduser()

    # Resolve relative to absolute path, if necessary
    p = p.resolve()

    # Create the data directory if necessary
    p.mkdir(exist_ok=True)

    # Attach the data path object to config
    config['dataPathObj'] = p

    # Set the project folder path
    config['projectsDirPathObj'] = p / 'projects'

    # Check if a config file is available, and, if so, load the config
    cp = p / 'config' / 'config.json'
    if cp.is_file():
        load_config(cp)

    # Check if global interface templates file is available, and, if so, load it
    gdit = p / 'global_templates' / 'interface_templates.json'
    if gdit.is_file():
        with gdit.open() as f:
            config['globalDefaultInterfaceTemplates'] = f.read()

    # Check if global project template file is available, and, if so, load it
    gdpt = p / 'global_templates' / 'project_template.json'
    if gdpt.is_file():
        with gdpt.open() as f:
            config['globalDefaultProjectTemplate'] = f.read()

    # Set up the data path scaffolding as needed
    scaffold_data_path()


""" Flask-User is a customizable user account management extension for Flask.

    :copyright: (c) 2013 by Ling Thio
    :author: Ling Thio (ling.thio@gmail.com)
    :license: Simplified BSD License, see LICENSE.txt for more details."""

from passlib.context import CryptContext
from flask import Blueprint, current_app, url_for, render_template
from flask_login import LoginManager, UserMixin as LoginUserMixin
from .db_adapters import DBAdapter
from .db_adapters import SQLAlchemyAdapter
from . import emails
from . import forms
from . import passwords
from . import settings
from . import tokens
from . import translations
from . import views
from . import signals
from .translations import get_translations

# Enable the following: from flask_user import current_user
from flask_login import current_user

# Enable the following: from flask_user import login_required, roles_required
from .decorators import *
# Enable the following: from flask_user import user_logged_in
from .signals import *


__version__ = '0.6.21'


def _call_or_get(function_or_property):
    return function_or_property() if callable(function_or_property) else function_or_property


def _flask_user_context_processor():
    """ Make 'user_manager' available to Jinja2 templates"""
    return dict(
        user_manager=current_app.user_manager,
        call_or_get=_call_or_get)

class UserManager(object):
    """ This is the Flask-User object that manages the User management process."""

    def __init__(self, db_adapter=None, app=None, **kwargs):
        """ Create the UserManager object """
        self.db_adapter = db_adapter
        self.app = app

        if db_adapter is not None and app is not None:
            self.init_app(app, db_adapter, **kwargs)


    def init_app(self, app, db_adapter=None,
                # Forms
                add_email_form=forms.AddEmailForm,
                change_password_form=forms.ChangePasswordForm,
                change_username_form=forms.ChangeUsernameForm,
                forgot_password_form=forms.ForgotPasswordForm,
                login_form=forms.LoginForm,
                register_form=forms.RegisterForm,
                resend_confirm_email_form=forms.ResendConfirmEmailForm,
                reset_password_form=forms.ResetPasswordForm,
                invite_form=forms.InviteForm,
                # Validators
                username_validator=forms.username_validator,
                password_validator=forms.password_validator,
                # View functions
                render_function=render_template,
                change_password_view_function=views.change_password,
                change_username_view_function=views.change_username,
                confirm_email_view_function=views.confirm_email,
                email_action_view_function=views.email_action,
                forgot_password_view_function=views.forgot_password,
                login_view_function=views.login,
                logout_view_function=views.logout,
                manage_emails_view_function=views.manage_emails,
                register_view_function=views.register,
                resend_confirm_email_view_function = views.resend_confirm_email,
                reset_password_view_function = views.reset_password,
                unconfirmed_email_view_function = views.unconfirmed,
                unauthenticated_view_function = views.unauthenticated,
                unauthorized_view_function = views.unauthorized,
                user_profile_view_function = views.user_profile,
                invite_view_function = views.invite,
                # Misc
                login_manager=LoginManager(),
                password_crypt_context=None,
                send_email_function = emails.send_email,
                make_safe_url_function = views.make_safe_url,
                token_manager=tokens.TokenManager(),
                legacy_check_password_hash=None
                ):
        """ Initialize the UserManager object """
        self.app = app
        if db_adapter is not None:
            self.db_adapter = db_adapter
        # Forms
        self.add_email_form = add_email_form
        self.change_password_form = change_password_form
        self.change_username_form = change_username_form
        self.forgot_password_form = forgot_password_form
        self.login_form = login_form
        self.register_form = register_form
        self.resend_confirm_email_form = resend_confirm_email_form
        self.reset_password_form = reset_password_form
        self.invite_form = invite_form
        # Validators
        self.username_validator = username_validator
        self.password_validator = password_validator
        # View functions
        self.render_function = render_function
        self.change_password_view_function = change_password_view_function
        self.change_username_view_function = change_username_view_function
        self.confirm_email_view_function = confirm_email_view_function
        self.email_action_view_function = email_action_view_function
        self.forgot_password_view_function = forgot_password_view_function
        self.login_view_function = login_view_function
        self.logout_view_function = logout_view_function
        self.manage_emails_view_function = manage_emails_view_function
        self.register_view_function = register_view_function
        self.resend_confirm_email_view_function = resend_confirm_email_view_function
        self.reset_password_view_function = reset_password_view_function
        self.unconfirmed_email_view_function = unconfirmed_email_view_function
        self.unauthenticated_view_function = unauthenticated_view_function
        self.unauthorized_view_function = unauthorized_view_function
        self.user_profile_view_function = user_profile_view_function
        self.invite_view_function = invite_view_function
        # Misc
        self.login_manager = login_manager
        self.token_manager = token_manager
        self.password_crypt_context = password_crypt_context
        self.send_email_function = send_email_function
        self.make_safe_url_function = make_safe_url_function
        self.legacy_check_password_hash = legacy_check_password_hash

        """ Initialize app.user_manager."""
        # Bind Flask-USER to app
        app.user_manager = self
        # Flask seems to also support the current_app.extensions[] list
        if not hasattr(app, 'extensions'):
            app.extensions = {}
        app.extensions['user'] = self

        # Set defaults for undefined settings
        settings.set_default_settings(self, app.config)

        # Make sure the settings are valid -- raise ConfigurationError if not
        settings.check_settings(self)

        # Initialize Translations -- Only if Flask-Babel has been installed
        if hasattr(app.jinja_env, 'install_gettext_callables'):
            app.jinja_env.install_gettext_callables(
                    lambda x: get_translations().ugettext(x),
                    lambda s, p, n: get_translations().ungettext(s, p, n),
                    newstyle=True)
        else:
            app.jinja_env.add_extension('jinja2.ext.i18n')
            app.jinja_env.install_null_translations()


        # Create password_crypt_context if needed
        if not self.password_crypt_context:
            self.password_crypt_context = CryptContext(
                    schemes=[app.config['USER_PASSWORD_HASH']])

        # Setup Flask-Login
        self.setup_login_manager(app)

        # Setup TokenManager
        self.token_manager.setup(app.config.get('SECRET_KEY'))

        # Add flask_user/templates directory using a Blueprint
        blueprint = Blueprint('flask_user', 'flask_user', template_folder='templates')
        app.register_blueprint(blueprint)

        # Add URL routes
        self.add_url_routes(app)

        # Add context processor
        app.context_processor(_flask_user_context_processor)

        # Prepare for translations
        _ = translations.gettext


    def setup_login_manager(self, app):

        # Flask-Login calls this function to retrieve a User record by user ID.
        # Note: user_id is a UNICODE string returned by UserMixin.get_id().
        # See https://flask-login.readthedocs.org/en/latest/#how-it-works
        @self.login_manager.user_loader
        def load_user_by_id(user_unicode_id):
            user_id = int(user_unicode_id)
            #print('load_user_by_id: user_id=', user_id)
            return self.get_user_by_id(user_id)

        self.login_manager.login_view = 'user.login'
        self.login_manager.init_app(app)


    def add_url_routes(self, app):
        """ Add URL Routes"""
        app.add_url_rule(self.login_url,  'user.login',  self.login_view_function,  methods=['GET', 'POST'])
        app.add_url_rule(self.logout_url, 'user.logout', self.logout_view_function, methods=['GET', 'POST'])
        if self.enable_confirm_email:
            app.add_url_rule(self.confirm_email_url, 'user.confirm_email', self.confirm_email_view_function)
            app.add_url_rule(self.resend_confirm_email_url, 'user.resend_confirm_email', self.resend_confirm_email_view_function, methods=['GET', 'POST'])
        if self.enable_change_password:
            app.add_url_rule(self.change_password_url, 'user.change_password', self.change_password_view_function, methods=['GET', 'POST'])
        if self.enable_change_username:
            app.add_url_rule(self.change_username_url, 'user.change_username', self.change_username_view_function, methods=['GET', 'POST'])
        if self.enable_forgot_password:
            app.add_url_rule(self.forgot_password_url, 'user.forgot_password', self.forgot_password_view_function, methods=['GET', 'POST'])
            app.add_url_rule(self.reset_password_url, 'user.reset_password', self.reset_password_view_function, methods=['GET', 'POST'])
        if self.enable_register:
            app.add_url_rule(self.register_url, 'user.register', self.register_view_function, methods=['GET', 'POST'])
        if self.db_adapter.UserEmailClass:
            app.add_url_rule(self.email_action_url,  'user.email_action',  self.email_action_view_function)
            app.add_url_rule(self.manage_emails_url, 'user.manage_emails', self.manage_emails_view_function, methods=['GET', 'POST'])
        app.add_url_rule(self.user_profile_url,  'user.profile',  self.user_profile_view_function,  methods=['GET', 'POST'])
        if self.enable_invitation:
            app.add_url_rule(self.invite_url, 'user.invite', self.invite_view_function, methods=['GET', 'POST'])
    # Obsoleted function. Replace with hash_password()
    def generate_password_hash(self, password):
        return passwords.hash_password(self, password)

    def hash_password(self, password):
        return passwords.hash_password(self, password)

    def get_password(self, user):
        use_auth_class = True if self.db_adapter.UserAuthClass and hasattr(user, 'user_auth') else False
        # Handle v0.5 backward compatibility
        if self.db_adapter.UserProfileClass:
            hashed_password = user.password
        else:
            hashed_password = user.user_auth.password if use_auth_class else user.password
        return hashed_password

    def update_password(self, user, hashed_password):
        use_auth_class = True if self.db_adapter.UserAuthClass and hasattr(user, 'user_auth') else False

        if use_auth_class:
            user.user_auth.password = hashed_password
        else:
            user.password = hashed_password
        self.db_adapter.commit()

    def verify_password(self, password, user):
        """
        Make it backward compatible to legacy password hash.
        In addition, if such password were found, update the user's password field.
        """
        verified = False
        hashed_password = self.get_password(user)

        try:
            verified = passwords.verify_password(self, password, hashed_password)
        except ValueError:
            legacy_check = self.legacy_check_password_hash
            if legacy_check:
                verified = legacy_check(hashed_password, password)
                if verified:
                    # update the hash
                    new_hash = self.hash_password(password)
                    self.update_password(user, new_hash)
        return verified

    def generate_token(self, user_id):
        return self.token_manager.generate_token(user_id)

    def verify_token(self, token, expiration_in_seconds):
        return self.token_manager.verify_token(token, expiration_in_seconds)

    def get_user_by_id(self, user_id):
        # Handle v0.5 backward compatibility
        ObjectClass = self.db_adapter.UserAuthClass if self.db_adapter.UserAuthClass and self.db_adapter.UserProfileClass else self.db_adapter.UserClass
        return self.db_adapter.get_object(ObjectClass, user_id)

    # NB: This backward compatibility function may be obsoleted in the future
    # Use 'get_user_by_id() instead.
    def find_user_by_id(self, user_id):
        print('Warning: find_user_by_id() will be deprecated in the future. User get_user_by_id() instead.')
        return self.get_user_by_id(user_id)

    def get_user_email_by_id(self, user_email_id):
        return self.db_adapter.get_object(self.db_adapter.UserEmailClass, user_email_id)

    # NB: This backward compatibility function may be obsoleted in the future
    # Use 'get_user_email_by_id() instead.
    def find_user_email_by_id(self, user_email_id):
        print('Warning: find_user_email_by_id() will be deprecated in the future. User get_user_email_by_id() instead.')
        return self.get_user_email_by_id(user_email_id)

    def find_user_by_username(self, username):
        user_auth = None

        # The username field can either be in the UserAuth class or in the User class
        if self.db_adapter.UserAuthClass and hasattr(self.db_adapter.UserAuthClass, 'username'):
            user_auth = self.db_adapter.ifind_first_object(self.db_adapter.UserAuthClass, username=username)

            # Handle v0.5 backward compatibility
            if self.db_adapter.UserProfileClass: return user_auth

            user = user_auth.user if user_auth else None
        else:
            user = self.db_adapter.ifind_first_object(self.db_adapter.UserClass, username=username)

        return user


    def find_user_by_email(self, email):
        user_email = None
        user_auth = None
        if self.db_adapter.UserEmailClass:
            user_email = self.db_adapter.ifind_first_object(self.db_adapter.UserEmailClass, email=email)
            user = user_email.user if user_email else None
        else:
            # The email field can either be in the UserAuth class or in the User class
            if self.db_adapter.UserAuthClass and hasattr(self.db_adapter.UserAuthClass, 'email'):
                user_auth = self.db_adapter.ifind_first_object(self.db_adapter.UserAuthClass, email=email)

                # Handle v0.5 backward compatibility
                if self.db_adapter.UserProfileClass: return (user_auth, user_email)

                user = user_auth.user if user_auth else None
            else:
                user = self.db_adapter.ifind_first_object(self.db_adapter.UserClass, email=email)

        return (user, user_email)

    def email_is_available(self, new_email):
        """ Return True if new_email does not exist.
            Return False otherwise."""
        user, user_email = self.find_user_by_email(new_email)
        return (user==None)

    def username_is_available(self, new_username):
        """ Return True if new_username does not exist or if new_username equals old_username.
            Return False otherwise."""
        # Allow user to change username to the current username
        if _call_or_get(current_user.is_authenticated):
            current_username = current_user.user_auth.username if self.db_adapter.UserAuthClass and hasattr(current_user, 'user_auth') else current_user.username
            if new_username == current_username:
                return True
        # See if new_username is available
        return self.find_user_by_username(new_username)==None

    def send_reset_password_email(self, email):
        # Find user by email
        user, user_email = self.find_user_by_email(email)
        if user:
            # Generate reset password link
            token = self.generate_token(int(user.get_id()))
            reset_password_link = url_for('user.reset_password', token=token, _external=True)

            # Send forgot password email
            emails.send_forgot_password_email(user, user_email, reset_password_link)

            # Send forgot_password signal
            signals.user_forgot_password.send(current_app._get_current_object(), user=user)


class UserMixin(LoginUserMixin):
    """ This class adds methods to the User model class required by Flask-Login and Flask-User."""

    def is_active(self):
        if hasattr(self, 'active'):
            return self.active
        else:
            return self.is_enabled


    def set_active(self, active):
        if hasattr(self, 'active'):
            self.active = active
        else:
            self.is_enabled = active


    def has_role(self, *specified_role_names):
        """ Return True if the user has one of the specified roles. Return False otherwise.

            has_roles() accepts a 1 or more role name parameters
                has_role(role_name1, role_name2, role_name3).

            For example:
                has_roles('a', 'b')
            Translates to:
                User has role 'a' OR role 'b'
        """

        # Allow developers to attach the Roles to the User or the UserProfile object
        if hasattr(self, 'roles'):
            roles = self.roles
        else:
            if hasattr(self, 'user_profile') and hasattr(self.user_profile, 'roles'):
                roles = self.user_profile.roles
            else:
                roles = None
        if not roles: return False

        # Translates a list of role objects to a list of role_names
        user_role_names = [role.name for role in roles]

        # Return True if one of the role_names matches
        for role_name in specified_role_names:
            if role_name in user_role_names:
                return True

        # Return False if none of the role_names matches
        return False


    def has_roles(self, *requirements):
        """ Return True if the user has all of the specified roles. Return False otherwise.

            has_roles() accepts a list of requirements:
                has_role(requirement1, requirement2, requirement3).

            Each requirement is either a role_name, or a tuple_of_role_names.
                role_name example:   'manager'
                tuple_of_role_names: ('funny', 'witty', 'hilarious')
            A role_name-requirement is accepted when the user has this role.
            A tuple_of_role_names-requirement is accepted when the user has ONE of these roles.
            has_roles() returns true if ALL of the requirements have been accepted.

            For example:
                has_roles('a', ('b', 'c'), d)
            Translates to:
                User has role 'a' AND (role 'b' OR role 'c') AND role 'd'"""

        # Allow developers to attach the Roles to the User or the UserProfile object
        if hasattr(self, 'roles'):
            roles = self.roles
        else:
            if hasattr(self, 'user_profile') and hasattr(self.user_profile, 'roles'):
                roles = self.user_profile.roles
            else:
                roles = None
        if not roles: return False

        # Translates a list of role objects to a list of role_names
        user_role_names = [role.name for role in roles]

        # has_role() accepts a list of requirements
        for requirement in requirements:
            if isinstance(requirement, (list, tuple)):
                # this is a tuple_of_role_names requirement
                tuple_of_role_names = requirement
                authorized = False
                for role_name in tuple_of_role_names:
                    if role_name in user_role_names:
                        # tuple_of_role_names requirement was met: break out of loop
                        authorized = True
                        break
                if not authorized:
                    return False                    # tuple_of_role_names requirement failed: return False
            else:
                # this is a role_name requirement
                role_name = requirement
                # the user must have this role
                if not role_name in user_role_names:
                    return False                    # role_name requirement failed: return False

        # All requirements have been met: return True
        return True


    # Flask-Login is capable of remembering the current user ID in the browser's session.
    # This function enables the user ID to be encrypted as a token.
    # See https://flask-login.readthedocs.org/en/latest/#remember-me
    def get_auth_token(self):
        token_manager = current_app.user_manager.token_manager
        user_id = int(self.get_id())
        token = token_manager.encrypt_id(user_id)
        #print('get_auth_token: user_id=', user_id, 'token=', token)
        return token


    def has_confirmed_email(self):
        db_adapter = current_app.user_manager.db_adapter

        # Handle multiple emails per user: Find at least one confirmed email
        if db_adapter.UserEmailClass:
            has_confirmed_email = False
            user_emails = db_adapter.find_all_objects(db_adapter.UserEmailClass, user_id=self.id)
            for user_email in user_emails:
                if user_email.confirmed_at:
                    has_confirmed_email = True
                    break

        # Handle single email per user
        else:
            has_confirmed_email = True if self.confirmed_at else False

        return has_confirmed_email



from flask import Flask, Blueprint, send_from_directory, request, render_template, render_template_string, abort
from flask_mail import Mail
from htmlmin.main import minify
from pathlib import Path
from pprint import pprint
import argparse
import logging
import shutil
import tempfile
import json

# Simplejson package is required in order to "ignore" NaN values and implicitly
# convert them into null values. RFC JSON spec left out NaN values, even though
# ES5 supports them (https://www.ecma-international.org/ecma-262/5.1/#sec-4.3.23).
# By default, Python "json" module will allow & json-encode NaN values, but the
# Chrome JS engine will throw an error when trying to parse them. Simplejson
# package, with ignore_nan=True, will implicitly convert NaN values into null
# values. Find "ignore_nan" here: https://simplejson.readthedocs.io/en/latest/
import simplejson

from . import models
from .api import downsampleFile, getProject, getProjectsPayload, loadProjects
from .patternset import getAssignmentsPayload
from .config import set_data_path, config, FlaskConfigClass

from .flask_user import current_user, login_required, UserManager, SQLAlchemyAdapter
from .flask_user.signals import user_sent_invitation, user_registered






##### IMPORTS FOR FEATURIZERS
# from sklearn.linear_model import LinearRegression

from datetime import datetime
import pandas as pd
import numpy as np
import pytz

###
### Featurizer functionality - disabled for release
###
# from .modules.featurization.example import ExampleFeaturizer
# from .modules.featurization.mean import MeanFeaturizer
# from .modules.featurization.std import StandardDeviationFeaturizer
# from functools import partial
#
# # Dict of featurizers indexed by ID
# featurizers = { f.id: f for f in [
#     ExampleFeaturizer(),
#     MeanFeaturizer(),
#     StandardDeviationFeaturizer(),
#     CoeffOfVariationFeaturizer(),
#     MADFeaturizer(),
#     NFeaturizer(),
#     MinFeaturizer(),
#     MaxFeaturizer(),
#     MedianFeaturizer(),
#     RangeFeaturizer(),
#     RangeRatioFeaturizer(),
#     DataDenFeaturizer(),
#     MaxGapFeaturizer(),
#     LRSlopeFeaturizer(),
#     RobustSlopeFeaturizer(),
# ]}
featurizers = {}


def createApp():
    # Instantiate the Flask web application class
    app = Flask(__name__, template_folder=str(config['codeRootPathObj'] / 'static' / 'www' / 'templates'))

    # Auto-reload templates
    app.jinja_env.auto_reload = True

    # Make the root web path available for templates
    @app.context_processor
    def inject_dict_for_all_templates():
        return {
            'rootWebPath': config['rootWebPath']
        }

    # Minify HTML when possible
    @app.after_request
    def response_minify(response):
        if response.content_type == u'text/html; charset=utf-8':
            response.set_data(minify(response.get_data(as_text=True), remove_comments=True))
        return response

    # Apply Flask configuration for Flask-User package
    app.config.from_object(FlaskConfigClass)
    app.config.update({
        'SECRET_KEY': config['secret_key'],
        'SQLALCHEMY_DATABASE_URI': f"sqlite:///{(config['dataPathObj'] / 'database' / 'db.sqlite')}",
        **config['mail'],
        'USER_CHANGE_PASSWORD_URL': config['rootWebPath'] + app.config['USER_CHANGE_PASSWORD_URL'],
        'USER_CHANGE_USERNAME_URL': config['rootWebPath'] + app.config['USER_CHANGE_USERNAME_URL'],
        'USER_CONFIRM_EMAIL_URL': config['rootWebPath'] + app.config['USER_CONFIRM_EMAIL_URL'],
        'USER_EMAIL_ACTION_URL': config['rootWebPath'] + app.config['USER_EMAIL_ACTION_URL'],
        'USER_FORGOT_PASSWORD_URL': config['rootWebPath'] + app.config['USER_FORGOT_PASSWORD_URL'],
        'USER_INVITE_URL': config['rootWebPath'] + app.config['USER_INVITE_URL'],
        'USER_LOGIN_URL': config['rootWebPath'] + app.config['USER_LOGIN_URL'],
        'USER_LOGOUT_URL': config['rootWebPath'] + app.config['USER_LOGOUT_URL'],
        'USER_MANAGE_EMAILS_URL': config['rootWebPath'] + app.config['USER_MANAGE_EMAILS_URL'],
        'USER_PROFILE_URL': config['rootWebPath'] + app.config['USER_PROFILE_URL'],
        'USER_REGISTER_URL': config['rootWebPath'] + app.config['USER_REGISTER_URL'],
        'USER_RESEND_CONFIRM_EMAIL_URL': config['rootWebPath'] + app.config['USER_RESEND_CONFIRM_EMAIL_URL'],
        'USER_RESET_PASSWORD_URL': config['rootWebPath'] + app.config['USER_RESET_PASSWORD_URL']
    })

    # Initialize the db models with the Flask app
    models.init_flask_app(app)

    # Initialize Flask-Mail
    mail = Mail(app)

    # Setup Flask-User
    db_adapter = SQLAlchemyAdapter(models.db, models.User, UserInvitationClass=models.UserInvitation)  # Select database adapter
    user_manager = UserManager(db_adapter, app)  # Init Flask-User and bind to app

    # You may use this code snippet below to create initial/new admin users
    # (since it is not possible through the interface). Alternatively, you
    # could modify the database.
    #
    with app.app_context():
        if not models.User.query.first():
            from getpass import getpass
            print("You must create an admin user.")
            fn = input("First name: ")
            ln = input("Last name: ")
            em = input("E-mail: ")
            pw = getpass(prompt="Password: ")
            u = models.User(
                first_name=fn,
                last_name=ln,
                email=em,
                active=True,
                password=user_manager.hash_password(pw),
            )
            u.roles.append(models.Role(name='admin'))
            models.db.session.add(u)
            models.db.session.commit()

    # Load projects
    with app.app_context():
        loadProjects()

    # Instantiate a file for realtime, in-memory usage (probably temporary)
    # TODO(gus): Refactor realtime
    #rtf = File(projparent=None)

    @user_registered.connect_via(app)
    def after_registered_hook(sender):
        sender.logger.info("USER REGISTERED")

    @user_sent_invitation.connect_via(app)
    def after_invitation_hook(sender):
        sender.logger.info("USER SENT INVITATION")

    ### NON-SECURE AREAS (NO LOGIN REQUIRED) ###

    # Map our static assets to be served
    app.register_blueprint(Blueprint('css', __name__, static_url_path=config['rootWebPath'] + '/css', static_folder=str(config['codeRootPathObj'] / 'static' / 'www' / 'css')))
    app.register_blueprint(Blueprint('fonts', __name__, static_url_path=config['rootWebPath'] + '/fonts', static_folder=str(config['codeRootPathObj'] / 'static' / 'www' / 'fonts')))
    app.register_blueprint(Blueprint('img', __name__, static_url_path=config['rootWebPath'] + '/img', static_folder=str(config['codeRootPathObj'] / 'static' / 'www' / 'img')))
    app.register_blueprint(Blueprint('js', __name__, static_url_path=config['rootWebPath'] + '/js', static_folder=str(config['codeRootPathObj'] / 'static' / 'www' / 'js')))

    @app.route('/favicon.ico')
    def favicon():
        return send_from_directory(str(config['codeRootPathObj'] / 'static' / 'www' / 'img' / 'favicons'), 'favicon.ico')

    ### SECURE AREAS (LOGIN REQUIRED) ###
    ### All methods below should have ###
    ### the @login_required decorator ###

    @app.route(config['rootWebPath']+'/bokeh.html')
    @login_required
    def bokeh():
        return send_from_directory('../www', 'bokeh.html')

    @app.route(config['rootWebPath'] + '/close_all_files', methods=['GET'])
    @login_required
    def close_all_files():
        projects = getProjectsPayload(current_user.id)

        for p in projects:
            proj = getProject(p['id'])
            for f in proj.files:
                try:
                    f.close()

                except Exception as e:
                    return app.response_class(
                            response=simplejson.dumps({'success': False}),
                            status=200,
                            mimetype='application/json'
                        )

        ### To be implemented here...
        return app.response_class(
            response=simplejson.dumps({'success': True}),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath'] + '/close_all_project_files', methods=['GET'])
    @login_required
    def close_all_project_files():
        # Parse parameters
        project_id = request.args.get('project_id', type=int)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            return app.response_class(
                response=simplejson.dumps({'success': False}),
                status=200,
                mimetype='application/json'
            )

        ### To be implemented here...
        for f in project.files:
            try:
                f.close()
            except Exception as e:
                print(e)
                return app.response_class(
                        response=simplejson.dumps({'success': False}),
                        status=200,
                        mimetype='application/json'
                    )

        ### This is the success response -- if there is an error you want to catch, just
        ### the same response below but with success=False!
        return app.response_class(
            response=simplejson.dumps({'success': True}),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath'] + '/close_project_file', methods=['GET'])
    @login_required
    def close_project_file():
        # Parse parameters
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            return app.response_class(
                response=simplejson.dumps({'success': False}),
                status=200,
                mimetype='application/json'
            )

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            return app.response_class(
                response=simplejson.dumps({'success': False}),
                status=200,
                mimetype='application/json'
            )

        ### To be implemented here...
        try:
            file.close()
        except Exception as e:
            return app.response_class(
                response=simplejson.dumps({'success':False}),
                status=200,
                mimetype='application/json'
            )

        ### This is the success response -- if there is an error you want to catch, just
        ### the same response below but with success=False!
        return app.response_class(
            response=simplejson.dumps({'success': True}),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath']+'/create_annotation', methods=['GET'])
    @login_required
    def create_annotation():

        # Parse parameters
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)
        left = getFloatParamOrNone('xl')
        right = getFloatParamOrNone('xr')
        top = getFloatParamOrNone('yt')
        bottom = getFloatParamOrNone('yb')
        seriesID = request.args.get('sid')
        label = request.args.get('label')
        pattern_id = request.args.get('pattern_id', type=int)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            return app.response_class(
                response=simplejson.dumps({'success': False}),
                status=200,
                mimetype='application/json'
            )

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            return app.response_class(
                response=simplejson.dumps({'success': False}),
                status=200,
                mimetype='application/json'
            )

        # Write the annotation
        newAnnotationID = file.createAnnotation(current_user.id, left, right, top, bottom, seriesID, label, pattern_id)
        def deleteAnnotations():
            allAnnotations = models.Annotation.query.filter_by(project_id=project_id).all()
            for annotation in allAnnotations:
                models.db.session.delete(annotation)
            models.db.session.commit()
        # deleteAnnotations()

        # Output response
        return app.response_class(
            response=simplejson.dumps({'success': True, 'id': newAnnotationID}),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath']+'/delete_annotation')
    @login_required
    def delete_annotation():

        # Parse parameters
        id = request.args.get('id', type=int)
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            return app.response_class(
                response=simplejson.dumps({'success': False}),
                status=200,
                mimetype='application/json'
            )

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            return app.response_class(
                response=simplejson.dumps({'success': False}),
                status=200,
                mimetype='application/json'
            )

        # Write the annotation
        deletionSuccess = file.deleteAnnotation(current_user.id, id)

        # Output response
        return app.response_class(
            response=simplejson.dumps({'success': deletionSuccess}),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath']+'/detect_patterns', methods=['GET'])
    @login_required
    def detect_patterns():

        # TODO(gus): Add checks here

        # Parse the series name and alert parameters
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)
        type = request.args.get('type')
        series = request.args.get('series')
        thresholdlow = request.args.get('thresholdlow', type=float) if request.args.get('thresholdlow') != '' else None
        thresholdhigh = request.args.get('thresholdhigh', type=float) if request.args.get('thresholdhigh') != '' else None
        duration = request.args.get('duration', type=float)
        persistence = request.args.get('persistence', type=float)/100
        maxgap = request.args.get('maxgap', type=float)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID '{project_id}' not found.")
            abort(404, description="Project not found.")
            return

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            return app.response_class(
                response=simplejson.dumps([]),
                status=200,
                mimetype='application/json'
            )

        # Run pattern detection
        alerts = file.detectPatterns(
            type=type,
            series=series,
            thresholdlow=thresholdlow,
            thresholdhigh=thresholdhigh,
            duration=duration,
            persistence=persistence,
            maxgap=maxgap
        )

        # Output response
        return app.response_class(
            response=simplejson.dumps(alerts, ignore_nan=True),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath'] + '/featurize')
    @login_required
    def featurize():

        # Parse parameters
        featurizer = request.args.get('featurizer')
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)
        series = request.args.get('series')
        left = request.args.get('left', type=float)
        right = request.args.get('right', type=float)
        params = json.loads(request.args.get('params'));

        print("\n".join([
            f"\nFeaturizer {featurizer} requested, with parameters:",
            f"Project ID: {project_id}",
            f"File ID: {file_id}",
            f"Series: {series}",
            f"Left: {left}",
            f"Right: {right}",
            f"Params: {params}\n",
        ]))

        if featurizer in featurizers:
            featurizerFunction = featurizers[featurizer].getFeaturizeFunction(params)
        else:
            raise Exception(f"Unknown featurizer requested: {featurizer}")

        try:

            # Get the project
            project = getProject(project_id)
            if project is None:
                logging.error(f"Project ID {project_id} not found.")
                abort(404, description="Project not found.")
                return

            # Get the file
            file = project.getFile(file_id)
            if file is None:
                logging.error(f"File ID {file_id} not found.")
                return app.response_class(
                    response=simplejson.dumps({'success': False}),
                    status=200,
                    mimetype='application/json'
                )

            utc = pytz.UTC

            s = file.getSeries(series)
            if s is None:
                raise Exception('series not found')
            df = s.getDataAsDF().set_index('time')
            window_size = params['window_size']

            # # Option 1 (sometimes this works) – make sure to correlate with option above
            left = datetime.fromtimestamp(left).astimezone(utc)
            right = datetime.fromtimestamp(right).astimezone(utc)

            # Option 2 (and sometimes this works) – make sure to correlate with option above
            # left = np.datetime64(datetime.fromtimestamp(left).astimezone(utc))
            # right = np.datetime64(datetime.fromtimestamp(right).astimezone(utc))

            # # This was from an earlier attempt, probably discard...
            # left = pd.to_datetime(datetime.fromtimestamp(left).astimezone(utc))
            # right = pd.to_datetime(datetime.fromtimestamp(right).astimezone(utc))

            df = df[(df.index >= left) & (df.index <= right)]

            # TODO: Might make label side configurable

            featurization = df.resample(window_size, label='right').agg(featurizerFunction).replace(np.inf, np.nan).replace(-np.inf, np.nan).dropna().reset_index()
            print(featurization)

            # Option 1 (sometimes this works) – make sure to correlate with option above
            featurization['time'] = ((featurization['time'].dt.tz_convert(utc) - pd.Timestamp("1970-01-01").replace(tzinfo=utc)) // pd.Timedelta("1ms")) / 1000

            # Option 2 (and sometimes this works) – make sure to correlate with option above
            # featurization['time'] = ((featurization['time'].dt.tz_localize('UTC') - pd.Timestamp("1970-01-01").replace(tzinfo=utc)) // pd.Timedelta("1ms")) / 1000

            nones = [None] * featurization.shape[0]
            data = [list(i) for i in zip(featurization['time'], nones, nones, featurization['value'])]
            response = {
                'id': f"{window_size}_{featurizer}_{series}",
                'success': True,
                'labels': ['Date/Offset', 'Min', 'Max', 'Sample Entropy for '+series],
                'data': data,
                'output_type': 'real',
                'temporary': True,
            }

            # Output response
            return app.response_class(
                response=simplejson.dumps(response, ignore_nan=True),
                status=200,
                mimetype='application/json',
            )

        except Exception as e:

            logging.error("There was an exception while featurizing.", exc_info=True)

            response = {
                'success': False,
                'error_message': f"There was an error while featurizing: {e}",
            }

            return app.response_class(
                response=simplejson.dumps(response, ignore_nan=True),
                status=200,
                mimetype='application/json',
            )

    @app.route(config['rootWebPath']+'/get_project_annotations')
    @login_required
    def get_project_annotations():

        # Parse parameters
        project_id = request.args.get('project_id', type=int)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            abort(404, description="Project not found.")
            return

        # Assemble project annotations
        projectAnnotations = project.getAnnotationsOutput(current_user.id)

        # Output response
        return app.response_class(
            response=simplejson.dumps(projectAnnotations, ignore_nan=True),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath']+'/')
    @app.route(config['rootWebPath']+'/index.html')
    @login_required
    def index():
        return render_template('index.html', projects=getProjectsPayload(current_user.id), assignments=getAssignmentsPayload(current_user.id))

    @app.route(config['rootWebPath']+'/initial_file_payload')
    @login_required
    def initial_file_payload():
        # Parse parameters
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            abort(404, description="Project not found.")
            return

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            abort(404, description="File not found.")
            return

        # Assemble the initial file payload (full zoomed-out & downsampled, if
        # necessary, datasets for all data series.
        initialFilePayload = file.getInitialPayload(current_user.id)

        # Output response
        return app.response_class(
            response=simplejson.dumps(initialFilePayload, ignore_nan=True),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath']+'/project')
    @login_required
    def project():

        # Parse parameters
        id = request.args.get('id', type=int)

        p = getProject(id)
        if p is None:
            logging.error(f"Project ID {id} not found.")
            abort(404, description="Project not found.")
            return

        # Project payload data for the HTML template
        projectPayload = p.getInitialPayload(current_user.id)

        # Assemble the data into a payload. We JSON-encode this twice. The first
        # one converts the dict into JSON. The second one essentially makes the
        # JSON string safe to drop straight into JavaScript code, as we are doing.
        projectPayloadJSON = simplejson.dumps(simplejson.dumps(projectPayload))








        #### TEMP FOR FEATURIZERS

        # Assemble the data into a payload. We JSON-encode this twice. The first
        # one converts the dict into JSON. The second one essentially makes the
        # JSON string safe to drop straight into JavaScript code, as we are doing.
        featurizersJSONPayload = simplejson.dumps(simplejson.dumps({
            f.id: {
                'id': f.id,
                'name': f.name,
                'fields': f.getFields(),
            } for f in featurizers.values()
        }))

        # m = MeanFeaturizer()
        #
        # # Assemble the data into a payload. We JSON-encode this twice. The first
        # # one converts the dict into JSON. The second one essentially makes the
        # # JSON string safe to drop straight into JavaScript code, as we are doing.
        # featurizersJSONPayload = simplejson.dumps(simplejson.dumps(m.getFields()))







        return render_template('project.html', project_name=projectPayload['project_name'], payload=projectPayloadJSON, featurizersJSONPayload=featurizersJSONPayload)

    @app.route(config['rootWebPath']+'/series_ranged_data', methods=['GET'])
    @login_required
    def series_ranged_data():

        # Parse parameters
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)
        series = request.args.getlist('s[]')
        start = request.args.get('start', type=float)
        stop = request.args.get('stop', type=float)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            abort(404, description="Project not found.")
            return

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            abort(404, description="File not found.")
            return

        # Assemble the series ranged data
        seriesRangedData = file.getSeriesRangedOutput(series, start, stop)

        # Output response
        return app.response_class(
            response=simplejson.dumps(seriesRangedData, ignore_nan=True),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath']+'/update_annotation', methods=['GET'])
    @login_required
    def update_annotation():

        # Parse parameters
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)
        id = request.args.get('id')
        left = getFloatParamOrNone('xl')
        right = getFloatParamOrNone('xr')
        top = getFloatParamOrNone('yt')
        bottom = getFloatParamOrNone('yb')
        seriesID = request.args.get('sid')
        label = request.args.get('label')

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            return app.response_class(
                response=simplejson.dumps({ 'success': False }),
                status=200,
                mimetype='application/json'
            )

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            return app.response_class(
                response=simplejson.dumps({ 'success': False }),
                status=200,
                mimetype='application/json'
            )

        # Update the annotation
        updateSuccess = file.updateAnnotation(current_user.id, id, left, right, top, bottom, seriesID, label)

        # Output the response
        return app.response_class(
            response=simplejson.dumps({'success': updateSuccess}),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath'] + '/user_manage')
    @login_required
    def user_manage():
        pprint(vars(current_user))
        return render_template_string("""
                {% extends "base.html" %}
                {% block content %}
                    <h2>Profile Page</h2>
                    <p>Hello {{ current_user.email }},</p>
                    <p><a href="{{ url_for('home_page') }}">Home Page</a></p>
                    <p><a href="{{ url_for('user.change_password') }}">Change password</a></p>
                    <p><a href="{{ url_for('user.invite') }}">Invite User</a></p>
                    <p><a href="{{ url_for('user.logout') }}">Sign out</a></p>
                {% endblock %}
                """)

    ### HELPERS ###

    # Returns the named request parameter as a float or None if the parameter is
    # empty or not present. The default return of None may be overridden.
    def getFloatParamOrNone(name, default=None):
        param = request.args.get(name)
        if param is not None and len(param) > 0:
            return request.args.get(name, type=float)
        else:
            return default

    return app

def main():

    # Parse command-line arguments
    parser = argparse.ArgumentParser(prog='python -m auviewer.serve', description='Auton Lab Universal Viewer')
    parser.add_argument('datapath', type=str, nargs='?', help='Path to data directory (may be empty if starting new)')
    parser.add_argument('-ds', '--downsample', metavar=('original_file', 'destination_path'), type=str, nargs=2, help='Downsample a single original file to a destination.')
    args = parser.parse_args()

    # Handle a downsample request
    if args.downsample is not None:
        print(f"Downsampling file {Path(args.downsample[0]).resolve()} to destination {Path(args.downsample[1]).resolve()}.")
        downsampleFile(args.downsample[0], args.downsample[1])
        return

    # Handle no data path or file argument
    if args.datapath is None:
        # If no argument provided, prompt user
        assumed_datapath = './auvdata'
        dp = input(f"AUViewer data path [{assumed_datapath}]: ") or assumed_datapath
        logging.info(f"Using data path {dp}")
        set_data_path(dp)

    # Handle single file argument
    elif Path(args.datapath).is_file():

        # If a file was provided, assume it's one-off view timeseries request
        logging.warning("Running viewer against a single file, using a temp directory for data storage (annotations, etc. will be lost).")

        # Establish paths we'll be working wtih
        temp_datapath = Path(tempfile.gettempdir()) / 'auvdata'
        filedir = temp_datapath / 'projects' / 'temp' / 'originals'
        gtdir = temp_datapath / 'global_templates'
        prjtmplt = gtdir / 'project_template.json'

        target_file = Path(args.datapath).resolve()
        symlink_file = filedir / target_file.name

        logging.info(f"Using temporary data path {temp_datapath}")

        # Delete the previous temp data folder if necessary
        if temp_datapath.is_dir():
            shutil.rmtree(temp_datapath)

        # Create temp data, project, and originals folders recursively
        filedir.mkdir(mode=0o700, parents=True, exist_ok=False)

        # Create global templates folder
        gtdir.mkdir(mode=0o700, exist_ok=False)

        # Create a global project template which shows all series by default
        with prjtmplt.open('x') as f:
            print('{ "series": { "_default": { "show": false } } }', file=f)

        # Establish symlink to the target file
        symlink_file.symlink_to(target_file)

        # Set datapath to our temp datapath
        set_data_path(temp_datapath)

    # Handle data path argument
    else:
        # Set the data path provided
        set_data_path(args.datapath)

    app = createApp()

    # Open auviewer in the browser, just before we spin up the server
    browser_url = f"http://{config['host']}{':' + str(config['port']) if str(config['port']) != '80' else ''}/{config['rootWebPath']}".rstrip('/')
    bannerMsgPrefix = '\033[96m\033[1m\033[4m'
    fmtEndSuffix = '\033[0m'
    if args.datapath is not None and Path(args.datapath).is_file():
        print('\033[93m'+"\nNOTE: AUViewer is running against a single file using a temporary directory for data storage. Any annotations, patterns, etc. will be lost!"+fmtEndSuffix)
        print(f"\n{bannerMsgPrefix}You may access your file at: {browser_url}/project?id=1#file_id=1\n{fmtEndSuffix}")
    else:
        print(f"\n{bannerMsgPrefix}You may access AUViewer at: {browser_url}\n{fmtEndSuffix}")

    app.run(host=config['host'], port=config['port'], debug=True, use_reloader=False)
    return app

# Start development web server
if __name__ == '__main__':
    app = main()



import os

from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext as _build_ext
from glob import glob

from auviewer import __VERSION__

NAME = 'auviewer'
VERSION = __VERSION__

RECOMPILE_CYTHON = True

# Handle optional cython re-compilation (should be disabled by default for
# standard client build-installs)
ext = '.pyx' if RECOMPILE_CYTHON else '.c'
extensions = [Extension("auviewer.cylib", ["auviewer/cylib"+ext])]
if RECOMPILE_CYTHON:
    from Cython.Build import cythonize
    from numpy import get_include
    extensions = cythonize(extensions, language_level=3)

def read(fn):
    return open(os.path.join(os.path.dirname(__file__), fn)).read()

pkg_files = [elem.replace('auviewer/', '') for elem in glob('auviewer/static/**/*', recursive=True) if os.path.isfile(elem)]

# Workaround to import numpy without assuming it's initially installed.
# From https://stackoverflow.com/questions/19919905/how-to-bootstrap-numpy-installation-in-setup-py
class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())

setup(
    name=NAME,
    version=VERSION,
    description='A general-purpose time series exploration & annotation tool.',
    long_description='AUViewer is a tool for viewing & annotating time series data, usable as a standalone tool for individual use or as a web-based application served to many users for annotation.',
    author='Gus Welter',
    author_email='gwelter@cmu.edu',
    maintainer='Gus Welter',
    maintainer_email='gwelter@cmu.edu',
    license='MIT',
    entry_points={
        'console_scripts': [
            'auv=auviewer.serve:main'
        ]
    },
    cmdclass={'build_ext':build_ext},
	ext_modules=extensions,
    package_data={NAME: pkg_files},
    install_requires=[
        'audata',
        'bcrypt',
        'email_validator',
        'flask==2.3.3',
        'flask-login',
        'flask-mail',
        'flask-sqlalchemy',
        'flask-wtf',
        'htmlmin',
        'jsbeautifier',
        'numpy>=1.19.5',
        'pandas',
        'passlib',
        'psutil',
        'pycryptodome',
        'simplejson',
        'sqlalchemy>=2.0'
    ],
    packages=find_packages(),
    setup_requires=['numpy'],
    python_requires='>=3.7',
    url='https://github.com/autonlab/auviewer',
    classifiers=[
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
    ],
)



from collections import deque
from copy import copy
from math import ceil
from threading import Lock
import logging
import numpy as np
import pandas as pd
import time
import psutil

from .config import config
from .rawdata import RawData
from .downsampleset import DownsampleSet

from .cylib import generateThresholdAlerts

# Represents a single time series of data.
class Series:

    # Reads a series into memory and builds the downsampling. Fileparent is a reference
    # to the File class instance which contains the Series.
    def __init__(self, ds, timecol, valcol, fileparent):

        # The series ID is the dataset path in the HDF5 file
        self.id = f'{ds.name}:{valcol}'

        # Holds the ordered (hierarchical) list of groups and, ultimately,
        # dataset to which the series belongs. So, this is like a folder path,
        # where the first element is the outermost group name and the last
        # element is the dataset name of the series.
        self.h5path = [e for e in ds.name.split('/') if len(e) > 0]

        # Holds the ordered (hierarchical) list of groups where downsamples will
        # be stored in the processed file.
        self.h5pathDownsample = copy(self.h5path)
        self.h5pathDownsample.append(valcol)

        # Holds the time & value column names in the HDF dataset
        self.timecol = timecol
        self.valcol = valcol

        # Holds a reference to the file parent which contains the series
        self.fileparent = fileparent

        # Initialize raw data in memory
        self.initializeRawDataInMemory()

        if self.fileparent.mode() == 'file':

            # Holds the raw data set
            self.rd = RawData(self)

            # Holds the downsample set
            self.dss = DownsampleSet(self)

            # Grab the unit, if available
            try:
                self.units = self.fileparent.f['/'.join(self.h5path)].meta['dwc_meta']['unitLabel']
            except:
                self.units = ""

            logging.debug(f"Units: {self.units}")

        elif self.fileparent.mode() == 'realtime':

            self.dequeLock = Lock()

    # Add data to the series (e.g. for realtime). The data is assumed to occur
    # after any existing data. The data parameter should be a dict of lists as
    # follows:
    #
    #   {
    #     'times': [ t1, t2, ... , tn ],
    #     'values': [ v1, v2, ... , vn ]
    #   }
    def addData(self, data):

        # ATW: TODO: Reimplement.
        raise Exception('Data appending is currently unsupported.')

        if not isinstance(data, dict) or not isinstance(data['times'], list) or not isinstance(data['values'], list):
            raise Exception('Invalid seriesData parameter received for series.addData().')

        # Add the new times & values.
        # TODO(gus): Once connected, check whether this is most efficient way to
        # join lists.
        # self.rawTimes = self.rawTimes + data['times']
        # self.rawValues = self.rawValues + data['values']
        with self.dequeLock:
            self.rawTimes.extend(data['times'])
            self.rawValues.extend(data['values'])

        return

    def generateThresholdAlerts(
            self, 
            thresholdlow, 
            thresholdhigh, 
            mode, duration, 
            persistence, 
            maxgap, 
            expected_frequency=0, 
            min_density=0, 
            drop_values_below=None, 
            drop_values_above=None,
            drop_values_between=None,
        ):

        # Pull raw data for the series into memory
        self.pullRawDataIntoMemory()

        # assemble a data numpy array with two columns: self.rawTimes and self.rawValues
        data = np.array([self.rawTimes, self.rawValues]).T

        # Drop values below the drop_values_below threshold
        if drop_values_below is not None:
            data = data[data[:,1] >= drop_values_below]

        # Drop values above the drop_values_above threshold
        if drop_values_above is not None:
            data = data[data[:,1] <= drop_values_above]

        # Drop values between the drop_values_between thresholds
        if drop_values_between is not None:
            data = data[(data[:,1] <= drop_values_between[0]) | (data[:,1] >= drop_values_between[1])]

        # Run through the data and generate alerts
        alerts = generateThresholdAlerts(data[:,0], data[:,1], thresholdlow, thresholdhigh, mode, duration, persistence, maxgap, ceil(expected_frequency*duration*min_density))

        # Remove raw data for the series fromm memory
        self.initializeRawDataInMemory()

        return alerts

    def getDataAsDF(self):
        """
        Returns the series data as a Pandas DataFrame, with columns time and value.
        Does not cache data in the class instance after returning.
        """
        return self.fileparent.f['/'.join(self.h5path)].get(datetimes=True)[[self.timecol, self.valcol]].rename(columns={self.timecol: 'time', self.valcol: 'value'})

    # Produces JSON output for the series at the maximum time range.
    def getFullOutput(self):

        logging.info("Assembling full output for " + self.id + ".")

        if self.fileparent.mode() == 'realtime':

            with self.dequeLock:
                nones = [None] * len(self.rawTimes)
                data = [list(i) for i in zip(self.rawTimes, nones, nones, self.rawValues)]
            output_type = 'real'

        elif self.fileparent.mode() == 'file':

            # Attempt to retrieve the full downsample output
            downsampleFullOutput = self.dss.getFullOutput()

            # Set data either to the retrieved downsample or to the raw data
            if downsampleFullOutput is not None:

                data = downsampleFullOutput.to_records(index=False).tolist()
                output_type = 'downsample'

            else:

                # Get reference to the series datastream from the HDF5 file
                rawTimes, rawValues = self.pullRawDataIntoMemory(returnValuesOnly=True)
                nones = [None] * len(rawTimes)
                data = [list(i) for i in zip(rawTimes, nones, nones, rawValues)]
                output_type = 'real'

        else:

            # Having reached this point, the mode is invalid.
            raise Exception('Invalid mode found for fileparent in series.getFullOutput():', self.fileparent.mode())

        logging.info(f"Completed assembly of full ({'downsampled' if output_type=='downsample' else 'raw'}) output for {self.id}.")

        # Return the JSON-ready output object
        return {
            "id": self.id,
            "labels": ['Date/Offset', 'Min', 'Max', simpleSeriesName(self.id)],
            "data": data,
            "output_type": output_type,
            "units": self.units
        }

    # Produces JSON output for the series over a specified time range, with
    # starttime and stoptime being time offset floats in seconds.
    def getRangedOutput(self, starttime, stoptime):

        logging.info(f"Assembling ranged output for {self.id}.")

        # Getting ranged output is not supported in realtime-mode.
        if self.fileparent.mode() == 'realtime':
            raise Exception('series.getRangedOutput() is not available in realtime-mode.')

        # Get the appropriate downsample for this time range
        ds = self.dss.getRangedOutput(starttime, stoptime)
        if isinstance(ds, pd.DataFrame):
            data = ds.to_records(index=False).tolist()
            output_type = 'downsample'

        
        # if (not isinstance(ds, pd.DataFrame) or pd.DataFrame.empty):
        else:
            data = self.rd.getRangedOutput(starttime, stoptime)
            output_type = 'real'
        # print(data)
        logging.info(f"Completed assembly of ranged ({'downsampled' if output_type=='downsample' else 'raw'}) output for {self.id}.")

        # Return the JSON-ready output object
        return {
            "id": self.id,
            "labels": ['Date/Offset', 'Min', 'Max', simpleSeriesName(self.id)],
            "data": data,
            "output_type": output_type,
            "units": self.units
        }

    # Process and store all downsamples for the series.
    def processAndStore(self):

        p = psutil.Process()

        logging.info(f"Processing & storing all downsamples for the series {self.id}")
        start = time.time()

        logging.info(f"MEM PRE-PULLD: {p.memory_full_info().uss / 1024 / 1024} MB")

        # Pull raw data for the series into memory
        try:
            self.pullRawDataIntoMemory()
        except Exception as e:
            logging.error(f"Error pulling raw data for series {self.id}. Raising exception.")
            raise e

        logging.info(f"MEM AFT-PULLD: {p.memory_full_info().uss/1024/1024} MB")

        # Build & store to file all downsamples for the series
        try:
            self.dss.processAndStore()
        except Exception as e:
            logging.error(f"Error processing & storing downsamples for series {self.id}. Raising exception.")
            raise e

        logging.info(f"MEM AFT-DSPRC: {p.memory_full_info().uss / 1024 / 1024} MB")

        # Remove raw data for the series fromm memory
        try:
            self.initializeRawDataInMemory()
        except Exception as e:
            logging.error(f"Error removing raw data for series {self.id}. Raising exception.")
            raise e

        logging.info(f"MEM AFT-REMVD: {p.memory_full_info().uss / 1024 / 1024} MB")

        end = time.time()
        logging.info(f"Completed processing & storing all downsamples for the series {self.id}. Took {round((end - start) / 60, 3)} minutes.")

    def pullRawDataIntoMemory(self, returnValuesOnly=False):
        """
        Pulls the raw data for the series from the file into memory (self.rawTimeOffsets and self.rawValues).
        If the returnValuesOnly is set, the function will return a tuple with the times & values and not hold
        them in the class instance.
        """

        logging.info(f"Reading raw series data into memory for {self.id}.")
        start = time.time()

        # If we're in realtime mode, this procedure is not applicable.
        if self.fileparent.mode() == 'realtime':
            logging.info("Reading raw series n/a since we're in mem mode. Returning.")
            return

        # Get reference to the series datastream from the HDF5 file
        dataset = self.fileparent.f['/'.join(self.h5path)][()]

        rawTimes = dataset[self.timecol].values.astype(np.float64)
        rawValues = dataset[self.valcol].values.astype(np.float64)

        # Drop nan values
        mask = ~np.isnan(rawValues)
        rawTimes = rawTimes[mask]
        rawValues = rawValues[mask]

        # Return the values if requested, otherwise attach them to the class instance.
        if returnValuesOnly:
            return rawTimes, rawValues
        else:
            self.rawTimes = rawTimes
            self.rawValues = rawValues

        end = time.time()
        logging.info(f"Finished reading raw series data into memory for {self.id} ({self.rawTimes.shape[0]} points). Took {round(end - start, 5)}s.")

    # Initializes the raw data stored for the series in memory and thereby
    # removes it from memory.
    def initializeRawDataInMemory(self):

        # Initialize raw data
        # self.rawTimes = []
        # self.rawValues = []
        self.rawTimes = deque(maxlen=config['M'])
        self.rawValues = deque(maxlen=config['M'])

def simpleSeriesName(s):
    simpleNameComponents = s.split('/')[-1].split(':')
    if simpleNameComponents[1] == 'value':
        return simpleNameComponents[0]
    else:
        return ':'.join(simpleNameComponents)



""" This file contains email sending functions for Flask-User.
    It uses Jinja2 to render email subject and email message. It uses Flask-Mail to send email.

    :copyright: (c) 2013 by Ling Thio
    :author: Ling Thio (ling.thio@gmail.com)
    :license: Simplified BSD License, see LICENSE.txt for more details."""

import smtplib
import socket
from flask import current_app, render_template

def _render_email(filename, **kwargs):
    # Render subject
    subject = render_template(filename+'_subject.txt', **kwargs)
    # Make sure that subject lines do not contain newlines
    subject = subject.replace('\n', ' ')
    subject = subject.replace('\r', ' ')
    # Render HTML message
    html_message = render_template(filename+'_message.html', **kwargs)
    # Render text message
    text_message = render_template(filename+'_message.txt', **kwargs)

    return (subject, html_message, text_message)

def send_email(recipient, subject, html_message, text_message):
    """ Send email from default sender to 'recipient' """

    # Disable email sending when testing
    if current_app.testing: return

    class SendEmailError(Exception):
        pass

    # Make sure that Flask-Mail has been installed
    try:
        from flask_mail import Message
    except:
        raise SendEmailError("Flask-Mail has not been installed. Use 'pip install Flask-Mail' to install Flask-Mail.")

    # Make sure that Flask-Mail has been initialized
    mail_engine = current_app.extensions.get('mail', None)
    if not mail_engine:
        raise SendEmailError('Flask-Mail has not been initialized. Initialize Flask-Mail or disable USER_SEND_PASSWORD_CHANGED_EMAIL, USER_SEND_REGISTERED_EMAIL and USER_SEND_USERNAME_CHANGED_EMAIL')

    try:

        # Construct Flash-Mail message
        message = Message(subject,
                recipients=[recipient],
                html = html_message,
                body = text_message)
        mail_engine.send(message)

    # Print helpful error messages on exceptions
    except (socket.gaierror, socket.error) as e:
        raise SendEmailError('SMTP Connection error: Check your MAIL_SERVER and MAIL_PORT settings.')
    except smtplib.SMTPAuthenticationError:
        raise SendEmailError('SMTP Authentication error: Check your MAIL_USERNAME and MAIL_PASSWORD settings.')

def get_primary_user_email(user):
    user_manager =  current_app.user_manager
    db_adapter = user_manager.db_adapter
    if db_adapter.UserEmailClass:
        user_email = db_adapter.find_first_object(db_adapter.UserEmailClass,
                user_id=int(user.get_id()),
                is_primary=True)
        return user_email
    else:
        return user


def send_confirm_email_email(user, user_email, confirm_email_link):
    # Verify certain conditions
    user_manager =  current_app.user_manager
    if not user_manager.enable_email: return
    if not user_manager.send_registered_email and not user_manager.enable_confirm_email: return

    # Retrieve email address from User or UserEmail object
    email = user_email.email if user_email else user.email
    assert(email)

    # Render subject, html message and text message
    subject, html_message, text_message = _render_email(
            user_manager.confirm_email_email_template,
            user=user,
            app_name=user_manager.app_name,
            confirm_email_link=confirm_email_link)

    # Send email message using Flask-Mail
    user_manager.send_email_function(email, subject, html_message, text_message)

def send_forgot_password_email(user, user_email, reset_password_link):
    # Verify certain conditions
    user_manager =  current_app.user_manager
    if not user_manager.enable_email: return
    assert user_manager.enable_forgot_password

    # Retrieve email address from User or UserEmail object
    email = user_email.email if user_email else user.email
    assert(email)

    # Render subject, html message and text message
    subject, html_message, text_message = _render_email(
            user_manager.forgot_password_email_template,
            user=user,
            app_name=user_manager.app_name,
            reset_password_link=reset_password_link)

    # Send email message using Flask-Mail
    user_manager.send_email_function(email, subject, html_message, text_message)

def send_password_changed_email(user):
    # Verify certain conditions
    user_manager =  current_app.user_manager
    if not user_manager.enable_email: return
    if not user_manager.send_password_changed_email: return

    # Retrieve email address from User or UserEmail object
    user_email = get_primary_user_email(user)
    assert(user_email)
    email = user_email.email
    assert(email)

    # Render subject, html message and text message
    subject, html_message, text_message = _render_email(
            user_manager.password_changed_email_template,
            user=user,
            app_name=user_manager.app_name)

    # Send email message using Flask-Mail
    user_manager.send_email_function(email, subject, html_message, text_message)

def send_registered_email(user, user_email, confirm_email_link):    # pragma: no cover
    # Verify certain conditions
    user_manager =  current_app.user_manager
    if not user_manager.enable_email: return
    if not user_manager.send_registered_email: return

    # Retrieve email address from User or UserEmail object
    email = user_email.email if user_email else user.email
    assert(email)

    # Render subject, html message and text message
    subject, html_message, text_message = _render_email(
            user_manager.registered_email_template,
            user=user,
            app_name=user_manager.app_name,
            confirm_email_link=confirm_email_link)

    # Send email message using Flask-Mail
    user_manager.send_email_function(email, subject, html_message, text_message)

def send_username_changed_email(user):  # pragma: no cover
    # Verify certain conditions
    user_manager =  current_app.user_manager
    if not user_manager.enable_email: return
    if not user_manager.send_username_changed_email: return

    # Retrieve email address from User or UserEmail object
    user_email = get_primary_user_email(user)
    assert(user_email)
    email = user_email.email
    assert(email)

    # Render subject, html message and text message
    subject, html_message, text_message = _render_email(
            user_manager.username_changed_email_template,
            user=user,
            app_name=user_manager.app_name)

    # Send email message using Flask-Mail
    user_manager.send_email_function(email, subject, html_message, text_message)

def send_invite_email(user, accept_invite_link):
    user_manager = current_app.user_manager
    if not user_manager.enable_email: return

    # Render subject, html message and text message
    subject, html_message, text_message = _render_email(
            user_manager.invite_email_template,
            user=user,
            app_name=user_manager.app_name,
            accept_invite_link=accept_invite_link)

    # Send email message using Flask-Mail
    user_manager.send_email_function(user.email, subject, html_message, text_message)



""" This file hashes and verifies passwords for Flask-User.

    :copyright: (c) 2013 by Ling Thio
    :author: Ling Thio (ling.thio@gmail.com)
    :license: Simplified BSD License, see LICENSE.txt for more details."""

from __future__ import print_function
import hashlib
import hmac
import base64

def generate_sha512_hmac(password_salt, password):
    """ Generate SHA512 HMAC -- for compatibility with Flask-Security """
    return base64.b64encode(hmac.new(password_salt, password.encode('utf-8'), hashlib.sha512).digest())

def hash_password(user_manager, password):
    """ Generate hashed password using SHA512 HMAC and the USER_PASSWORD_HASH hash function."""
    # Handle plaintext storage
    if user_manager.password_hash == 'plaintext':
        return password
    # Generate SHA512 HMAC -- For compatibility with Flask-Security
    if user_manager.password_hash_mode == 'Flask-Security':
        password = generate_sha512_hmac(user_manager.password_salt, password)
    # Use passlib to hash password
    hashed_password = user_manager.password_crypt_context.encrypt(password)

    return hashed_password

def verify_password(user_manager, password, hashed_password):
    """ Verify password with previously hashed password.
        Returns True on matching password.
        Returns False otherwise."""
    # Handle plaintext storage
    if user_manager.password_hash == 'plaintext':
        return password==hashed_password
    # Generate SHA512 HMAC -- For compatibility with Flask-Security
    if user_manager.password_hash_mode == 'Flask-Security':
        password = generate_sha512_hmac(user_manager.password_salt, password)

    return user_manager.password_crypt_context.verify(password, hashed_password)

